searchNodes=[{"doc":"A websocket client for Phoenix channelsSlipstream is a websocket client for connection to Phoenix.Channels. Slipstream is a bit different from existing websocket implementations in that:it's backed by :gun instead of :websocket_clientit has an await_* interface for performing actions synchronouslysmart retry strategies for reconnection and rejoining work out-of-the-boxa testing framework for clientshigh-level and low-level instrumentation with :telemetryBasic UsageThe intended use for Slipstream is to write asynchronous, callback-oriented GenServer-like modules that define socket clients. This approach makes it easy to write socket clients that resemble state-machines. A minimalistic example usage might be like so:defmodule MyApp.MySocketClient do @moduledoc &quot;&quot;&quot; A socket client for connecting to that other Phoenix server Periodically sends pings and asks the other server for its metrics. &quot;&quot;&quot; use Slipstream, restart: :temporary require Logger @topic &quot;backend-service:money-server&quot; def start_link(args) do Slipstream.start_link(__MODULE__, args, name: __MODULE__) end @impl Slipstream def init(config) do {:ok, connect!(config), {:continue, :start_ping}} end @impl Slipstream def handle_continue(:start_ping, socket) do timer = :timer.send_interval(self(), :request_metrics) {:noreply, assign(socket, :ping_timer, timer)} end @impl Slipstream def handle_connect(socket) do {:ok, join(socket, @topic)} end @impl Slipstream def handle_join(@topic, _join_response, socket) do # an asynchronous push with no reply: push(socket, @topic, &quot;hello&quot;, %{}) {:ok, socket} end @impl Slipstream def handle_info(:request_metrics, socket) do # we will asynchronously receive a reply and handle it in the # handle_reply/3 implementation below {:ok, ref} = push(socket, @topic, &quot;get_metrics&quot;, %{format: &quot;json&quot;}) {:noreply, assign(socket, :metrics_request, ref)} end @impl Slipstream def handle_reply(ref, metrics, socket) do if ref == socket.assigns.metrics_request do :ok = MyApp.MetricsPublisher.publish(metrics) end {:ok, socket} end @impl Slipstream def handle_message(@topic, event, message, socket) do Logger.error( &quot;Was not expecting a push from the server. Heard: &quot; &lt;&gt; inspect({@topic, event, message}) ) {:ok, socket} end @impl Slipstream def handle_disconnect(_reason, socket) do :timer.cancel(socket.assigns.ping_timer) {:stop, :normal, socket} end endSynchronicitySlipstream is designed to work asynchronously by default. Requests such as connect/2, join/3, and push/4 are asynchronous requests. When the remote server replies, the associated callback will be invoked (handle_connect/1, handle_join/3, and handle_reply/3 in cases of success, respectively). For all of these operations, though, you may await the outcome of the asynchronous request with await_* functions. E.g.iex&gt; {:ok, ref} = push(socket, &quot;room:lobby&quot;, &quot;msg:new&quot;, %{user: 1, msg: &quot;foo&quot;}) iex&gt; {:ok, %{&quot;created&quot; =&gt; true}} = await_reply(ref)Note that all await_* functions must be called from the slipstream process that emitted the request, or else they will timeout.While Slipstream provides a rich toolset for synchronicity, the asynchronous, callback-based workflow is recommended.GenServer operationsNote that Slipstream is in many ways a simple wrapper around a GenServer. As such, all GenServer functionality is possible with Slipstream clients, such as Kernel.send/2 or GenServer.call/3. For example, assume you have a slipstream client written like so:defmodule MyClient do use Slipstream require Logger def start_link(args) do Slipstream.start_link(__MODULE__, args, name: __MODULE__) end @impl Slipstream def init(config), do: connect(config) @impl Slipstream def handle_cast(:ping, socket) do Logger.info(&quot;pong&quot;) {:noreply, socket} end @impl Slipstream def handle_info(:hello, socket) do Logger.info(&quot;hello&quot;) {:noreply, socket} end @impl Slipstream def handle_call(:foo, _from, socket) do {:reply, {:ok, :bar}, socket} end .. endThis MyClient client is a GenServer, so the following are valid ways to interact with MyClient:iex&gt; GenServer.cast(MyClient, :ping) [info] pong :ok iex&gt; MyClient |&gt; GenServer.whereis |&gt; send(:hello) [info] hello :hello iex&gt; GenServer.call(MyClient, :foo) {:ok, :bar}Retry MechanismsSlipstream emulates the official phoenix.js package with its reconnection and re-join features. Slipstream.Configuration allows configuration of the back-off times with the :reconnect_after_msec and :rejoin_after_msec lists, respectively.To take advantage of these built-in mechanisms, a client must be written in the asynchronous GenServer-like manner and must use the reconnect/1 and rejoin/3 functions in its Slipstream.handle_disconnect/2 and Slipstream.handle_topic_close/3 callbacks, respectively. Note that the default implementation of these callbacks invokes these functions, so a client which does not explicitly define these callbacks will retry connection and joins.Take care to handle the :left case of Slipstream.handle_topic_close/3. In the case that a client attempts to leave a topic with leave/2, the callback will be invoked with a reason of :left. The default implementation of Slipstream.handle_topic_close/3 makes this distinction and simply no-ops on channel leaves.defmodule MyClientWithRetry do use Slipstream def start_link(config) do Slipstream.start_link(__MODULE__, config, name: __MODULE__) end @impl Slipstream def init(config), do: connect(config) @impl Slipstream def handle_connect(socket) do {:ok, join(socket, &quot;rooms:lobby&quot;, %{user_id: 1})} end @impl Slipstream def handle_disconnect(_reason, socket) do reconnect(socket) end @impl Slipstream def handle_topic_close(_topic, :left, socket) do {:ok, socket} end def handle_topic_close(topic, _reason, socket) do rejoin(socket, topic) end end","ref":"Slipstream.html","title":"Slipstream","type":"behaviour"},{"doc":"Declares that a module is a Slipstream socket clientSlipstream provides a use Slipstream macro that behaves similar to GenServer's use GenServer. This does a few things:a default implementation of child_spec/1, which is used to start the module as a GenServer. This may be overridden.imports for all documented functions in Slipstream and Slipstream.Socketa GenServer.handle_info/2 function clause which matches incoming events from the connection process and dispatches them to the various Slipstream callbacksa GenServer.handle_info/2 function clause which matches Slipstream commands. This is used to implement back-off retry mechanisms for reconnect/1 and rejoin/3.This provides a familiar and sleek interface for the common case of using Slipstream: an asynchronous callback-based GenServer module.It's not required to use this macro, though. Slipstream can be used in synchronous mode (via the await_* family of functions).OptionsAny options passed as the opts argument to __using__/1 are passed along to Supervisor.child_spec/2, as is done in the default child_spec/1 implementation for GenServer. Most notably, you may control the restart strategy of the client with the :restart option. See the example below.Examplesdefmodule MyApp.MySocketClient do # one crash/shutdown/exit will permanently terminate the server use Slipstream, restart: :temporary def start_link(args) do Slipstream.start_link(__MODULE__, args) end .. end","ref":"Slipstream.html#__using__/1","title":"Slipstream.__using__/1","type":"macro"},{"doc":"Awaits a pending connection request synchronously","ref":"Slipstream.html#await_connect/2","title":"Slipstream.await_connect/2","type":"function"},{"doc":"Awaits a pending connection request synchronously, raising on failure","ref":"Slipstream.html#await_connect!/2","title":"Slipstream.await_connect!/2","type":"function"},{"doc":"Awaits a pending disconnection request synchronously","ref":"Slipstream.html#await_disconnect/2","title":"Slipstream.await_disconnect/2","type":"function"},{"doc":"Awaits a pending disconnection request synchronously, raising on failure","ref":"Slipstream.html#await_disconnect!/2","title":"Slipstream.await_disconnect!/2","type":"function"},{"doc":"Awaits a pending join request synchronously","ref":"Slipstream.html#await_join/3","title":"Slipstream.await_join/3","type":"function"},{"doc":"Awaits a join request synchronously, raising on failure","ref":"Slipstream.html#await_join!/2","title":"Slipstream.await_join!/2","type":"function"},{"doc":"Awaits a leave request synchronously","ref":"Slipstream.html#await_leave/3","title":"Slipstream.await_leave/3","type":"function"},{"doc":"Awaits a leave request synchronously, raising on failure","ref":"Slipstream.html#await_leave!/2","title":"Slipstream.await_leave!/2","type":"function"},{"doc":"Awaits the server's message pushNote that unlike the other await_* functions, await_message/4 is a macro. This allows an author to match on patterns in the topic, event, and/or payload parts of a message.Examplesiex&gt; event = &quot;msg:new&quot; iex&gt; await_message(&quot;rooms:lobby&quot;, ^event, %{&quot;user_id&quot; =&gt; 5}) {:ok, &quot;rooms:lobby&quot;, &quot;msg:new&quot;, %{&quot;user_id&quot; =&gt; 5, body: &quot;hello&quot;}}","ref":"Slipstream.html#await_message/4","title":"Slipstream.await_message/4","type":"macro"},{"doc":"Awaits the server's message push, raising on failureSee await_message/4","ref":"Slipstream.html#await_message!/4","title":"Slipstream.await_message!/4","type":"macro"},{"doc":"Awaits the server's response to a message","ref":"Slipstream.html#await_reply/2","title":"Slipstream.await_reply/2","type":"function"},{"doc":"Awaits the server's response to a message, exiting on timeoutSee await_reply/2 for more information.","ref":"Slipstream.html#await_reply!/2","title":"Slipstream.await_reply!/2","type":"function"},{"doc":"Requests that the connection process undergoe garbage collectionIf you're using Slipstream to send large messages, you may wish to flush the process memory of the connection process between large messages. This can be achieved through garbage collection.Examplesiex&gt; collect_garbage(socket) :ok","ref":"Slipstream.html#collect_garbage/1","title":"Slipstream.collect_garbage/1","type":"function"},{"doc":"Requests connection to the remote endpointopts are passed to Slipstream.Configuration.validate/1 before sending.Note that this request for connection is asynchronous. A return value of {:ok, socket} does not mean that a connection has successfully been established.Examples{:ok, socket} = connect(uri: &quot;ws://localhost:4000/socket/websocket&quot;)","ref":"Slipstream.html#connect/2","title":"Slipstream.connect/2","type":"function"},{"doc":"Same as connect/2 but raises on configuration validation errorNote that connect!/2 will not necessarily raise an error on failure to connect. The ! only pertains to the potential for raising when the configuration is invalid.","ref":"Slipstream.html#connect!/2","title":"Slipstream.connect!/2","type":"function"},{"doc":"Requests that an open connection be closedThis function will no-op when the socket is not currently connected to any remote websocket server.Note that you do not need to use disconnect/1 to clean up a connection. The connection process monitors the slipstream client process and will shut down when it detects that the process has terminated.Disconnection may be awaited synchronously with await_disconnect/2Examples@impl Slipstream def handle_info(:chaos_monkey, socket) do {:ok, socket} = socket |&gt; disconnect() |&gt; await_disconnect() {:noreply, reconnect(socket)} end","ref":"Slipstream.html#disconnect/1","title":"Slipstream.disconnect/1","type":"function"},{"doc":"Invoked when a slipstream process receives a GenServer callBehaves the same as GenServer.handle_call/3","ref":"Slipstream.html#c:handle_call/3","title":"Slipstream.handle_call/3","type":"callback"},{"doc":"Invoked when a slipstream process receives a GenServer castBehaves the same as GenServer.handle_cast/2","ref":"Slipstream.html#c:handle_cast/2","title":"Slipstream.handle_cast/2","type":"callback"},{"doc":"Invoked when a connection has been established to a websocket serverThis callback provides a good place to join/3.Examples@impl Slipstream def handle_connect(socket) do {:noreply, socket} end","ref":"Slipstream.html#c:handle_connect/1","title":"Slipstream.handle_connect/1","type":"callback"},{"doc":"Invoked as a continuation of another GenServer callbackGenServer callbacks may end with signatures that declare that the next function invoked should be a continuation. E.g.def init(state) do {:ok, state, {:continue, :my_continue}} end # this will be invoked immediately after `init/1` def handle_continue(:my_continue, state) do # do something with state {:norelpy, state} endThis provides a way to schedule work to occur immediately after successful initialization or to break work across multiple callbacks, which can be useful for clients which are state-machine-like.See GenServer.handle_continue/2 for more information.","ref":"Slipstream.html#c:handle_continue/2","title":"Slipstream.handle_continue/2","type":"callback"},{"doc":"Invoked when a connection has been terminatedThe default implementation of this callback requests reconnectionExamples@impl Slipstream def handle_disconnect(_reason, socket) do {:ok, reconnect(socket)} end","ref":"Slipstream.html#c:handle_disconnect/2","title":"Slipstream.handle_disconnect/2","type":"callback"},{"doc":"Invoked when a slipstream process receives a messageBehaves the same as GenServer.handle_info/2","ref":"Slipstream.html#c:handle_info/2","title":"Slipstream.handle_info/2","type":"callback"},{"doc":"Invoked when the websocket server replies to the request to joinExamples@impl Slipstream def handle_join(&quot;rooms:echo&quot;, %{}, socket) do push(socket, &quot;rooms:echo&quot;, &quot;echo&quot;, %{&quot;ping&quot; =&gt; 1}) {:ok, socket} end","ref":"Slipstream.html#c:handle_join/3","title":"Slipstream.handle_join/3","type":"callback"},{"doc":"Invoked when a message is received on the websocket connectionThis callback will not be invoked for a message which is a reply. Those messages will be handled in handle_reply/3.Note that while replying is supported on the server-side of the Phoenix Channel protocol, it is not supported by a client. Messages sent from the server cannot be directly replied to.Examples@impl Slipstream def handle_message(&quot;room:lobby&quot;, &quot;new:msg&quot;, params, socket) do MyApp.Msg.create(params) {:ok, socket} end","ref":"Slipstream.html#c:handle_message/4","title":"Slipstream.handle_message/4","type":"callback"},{"doc":"Invoked when a message is received on the websocket connection which references a push from this client process.ref is the string reference returned from the push/2 which resulted in this reply.Examples@impl Slipstream def handle_join(topic, _params, socket) do my_req = push(socket, topic, &quot;msg:new&quot;, %{&quot;foo&quot; =&gt; &quot;bar&quot;}) {:ok, assign(socket, :request, my_req)} end @impl Slipstream def handle_reply(ref, reply, %{assigns: %{request: ref}} = socket) do IO.inspect(reply, label: &quot;reply to my request&quot;) {:ok, socket} end","ref":"Slipstream.html#c:handle_reply/3","title":"Slipstream.handle_reply/3","type":"callback"},{"doc":"Invoked when a join has concludedThis callback will be invoked in a few cases:the remote Phoenix.Channel crashes, e.g. by a raised errorthe client successfully leaves the topic with leave/2In the case that the client has left the topic, reason will simply be :left. If the remote channel crashes, the reason will be an error tuple {:error, params :: json_serializable()} where params is the message sent from the remote channel on chrash.The default implementation of this callback attempts to re-join the last-joined topic whenever reason != :left. If the reason is :left, the default implementation will no-op by returning {:ok, socket}.Examples@impl Slipstream def handle_topic_close(topic, _reason, socket) do {:ok, socket} = rejoin(socket, topic) end","ref":"Slipstream.html#c:handle_topic_close/3","title":"Slipstream.handle_topic_close/3","type":"callback"},{"doc":"Invoked when the slipstream process in startedBehaves the same as GenServer.init/1, but the return state must be a new Slipstream.Socket.t/0. Values from init/1 that you'd like to keep in state can be stored with Slipstream.Socket.assign/3.This callback is a good place to request connection with connect/2. Note that connect/2 is an asynchronous request for connection. Awaiting connection with await_connect/2 is unwise in many scenarios, however, because failure to connect may result in an exit from the process, crashing the supervision tree that started the process. If you wish to connect synchronously upon init, a better approach could be:@impl Slipstream def init(_args) do config = Application.fetch_env!(:my_app, __MODULE__) socket = new_socket() |&gt; assign(:connect_config, config) {:ok, socket, {:continue, :connect}} end @impl Slipstream def handle_continue(:connect, socket) do {:ok, socket} = connect(socket, socket.assigns.connect_config) {:ok, socket} = await_connect(socket) {:noreply, socket} endBut a more minimalistic approach that still provides safety in cases of configuration validation failures would be:defmodule MySocketClient do use Slipstream def start_link(args) do Slipstream.start_link(__MODULE__, args, name: __MODULE__) end @impl Slipstream def init(_args) do config = Application.fetch_env!(:my_app, __MODULE__) case connect(config) do {:ok socket} -&gt; {:ok, socket} {:error, reason} -&gt; Logger.error(&quot;Could not start \#{__MODULE__} because of &quot; &lt;&gt; &quot;validation failure: \#{inspect(reason)}&quot;) :ignore ned end .. endThe configuration could be stored in application config:# config/&lt;env&gt;.exs config :my_app, MySocketClient, uri: &quot;ws://example.org/socket/websocket&quot;, reconnect_after_msec: [200, 500, 1_000, 2_000]And in cases where the configuration validation fails, the MySocketClient process will not crash the application's supervision tree.Examples@impl Slipstream def init(config) do connect(config) end","ref":"Slipstream.html#c:init/1","title":"Slipstream.init/1","type":"callback"},{"doc":"Requests that a topic be joined in the current connectionMultiple topics may be joined by one Slipstream client, but each topic may only be joined once. Despite this, join/3 may be called on the same topic multiple times, but the result will be idempotent. The client will not request to join unless it has not yet joined that topic. In cases where you wish to begin a new session with a topic, you must first leave/2 and then join/3 again.The request to join will not error-out if the client is not connected to a remote server. In that case, the join/3 function will act as a no-op.A join can be awaited in a blocking fashion with await_join/3.Examples@impl Slipstream def handle_connect(socket) do {:ok, join(socket, &quot;rooms:lobby&quot;, %{user: 1})} end","ref":"Slipstream.html#join/3","title":"Slipstream.join/3","type":"function"},{"doc":"Requests that the given topic be leftNote that like joining, leaving is an asynchronous request and can be awaited with await_leave/3.Also similar to join/3, leave/2 is idempotent and will not raise an error if the provided topic is not currently joined.Examplesiex&gt; {:ok, socket} = leave(socket, &quot;room:lobby&quot;) |&gt; await_leave(&quot;rooms:lobby&quot;) iex&gt; join(socket, &quot;rooms:specific&quot;)","ref":"Slipstream.html#leave/2","title":"Slipstream.leave/2","type":"function"},{"doc":"Creates a new socket without immediately connecting to a remote websocketThis can be useful if you do not wish to request connection with connect/2 during the init/1 callback (because the init/1 callback requires that you return a Slipstream.Socket.t/0).Examplesdefmodule MySocketClient do use Slipstream .. @impl Slipstream def init(args) do {:ok, new_socket() |&gt; assign(:init_args, args)} end .. end iex&gt; new_socket() #Slipstream.Socket&lt;assigns: %{}, ...&gt;","ref":"Slipstream.html#new_socket/0","title":"Slipstream.new_socket/0","type":"function"},{"doc":"Requests that a message be pushed on the websocket connectionA channel has the ability to reply directly to a message, but this reply is asynchronous. Handle replies using the handle_reply/3 callback or by awaiting them synchronously with await_reply/2.Although this request to the remote server is asynchronous, the call to the transport process to transmit the push is synchronous and will exert back-pressure on calls to push/4, as push/4 blocks until the message has been sent by the transport.If you are pushing especially large messages, you may need to adjust the timeout argument so that the GenServer call does not exit with :timeout. The default value is 5_000 msec (5 seconds).A phoenix channel may decide to reply to a message sent with push/2. In order to link push requests to their replies, store the ref string returned from the call to push/4 and match on it in handle_reply/3.Examples@impl Slipstream def handle_join(:success, _response, state) do {:ok, hello_request} = push(socket, &quot;rooms:lobby&quot;, &quot;new:msg&quot;, %{user: 1, body: &quot;hello&quot;}) {:ok, Map.put(state, :hello_request, hello_request)} end @impl Slipstream def handle_reply(ref, reply, %{hello_request: ref} = state) do IO.inspect(reply, label: &quot;nice, a response.&quot;) {:ok, state} end","ref":"Slipstream.html#push/5","title":"Slipstream.push/5","type":"function"},{"doc":"Pushes, raising if the topic is not joined or if the channel is deadSame as push/4, but raises in cases of failure.This can be useful for pipeing into await_reply/2Examplesiex&gt; {:ok, result} = push!(socket, &quot;rooms:lobby&quot;, &quot;msg:new&quot;, params) |&gt; await_reply() {:ok, %{&quot;created?&quot; =&gt; true}}","ref":"Slipstream.html#push!/4","title":"Slipstream.push!/4","type":"function"},{"doc":"Request reconnection given the last-used connection configurationNote that when reconnect/1 is used to re-connect instead of connect/2 (or connect!/2), the slipstream process will attempt to reconnect with a retry mechanism with backoff. The process will wait an interval between reconnection attempts following the list of milliseconds provided in the :reconnect_after_msec key of configuration passed to connect/2 (or connect!/2).The handle_disconnect/2 callback will be invoked for each failure to re-connect, however, so an implementation of that callback which will simply retry with backoff can be achieved like so:@impl Slipstream def handle_disconnect(_reason, socket) do {:ok, socket} = reconnect(socket) endreconnect/1 may return an :error tuple in the case that the socket passed does not contain any connection information (which is added to the socket with connect/2 or connect!/2), or if the socket is currently connected. For a reconnect/1 call without configuration, the return pattern is {:error, :no_config}, and for a socket that is already connected, the pattern is {:error, :connected}.A reconnect may be awaited with await_connect/2.","ref":"Slipstream.html#reconnect/1","title":"Slipstream.reconnect/1","type":"function"},{"doc":"Requests that the specified topic be joined againIf params is not provided, the previously used value will be sent.In the case that the specified topic has not been joined before, rejoin/3 will return {:error, :never_joined}.Note that a rejoin may be awaited with await_join/3.Dealing with crashesWhen attempting to re-join a disconnected topic with rejoin/3, the Slipstream process will attempt to use backoff governed by the :rejoin_after_msec list in configuration passed to connect/2 (or connect!/2).The handle_topic_close/3 callback will be invoked with the for each crash, however, so a minimal implementation of that callback which achieves the backoff retry is like so:Examples@impl Slipstream def handle_topic_close(topic, _reason, socket) do {:ok, _socket} = rejoin(socket, topic) end","ref":"Slipstream.html#rejoin/3","title":"Slipstream.rejoin/3","type":"function"},{"doc":"Starts a slipstream client processThis function delegates to GenServer.start_link/3, so all options passable to that function are passable here as well. Most notably, you may name your slipstream clients using the same naming rules as GenServers, as in the example below.Examplesdefmodule MySlipstreamClient do use Slipstream def start_link(args) do Slipstream.start_link(__MODULE__, args, name: __MODULE__) end .. end","ref":"Slipstream.html#start_link/3","title":"Slipstream.start_link/3","type":"function"},{"doc":"Invoked when a slipstream process is terminatedNote that this callback is not always invoked as the process shuts down. See GenServer.terminate/2 for more information.It is wise to disconnect/1 in this callback (and such is the default implementation). This will gracefully end the websocket connection. This is the behavior of the default implementation of terminate/2.Examples@impl Slipstream def terminate(reason, socket) do Logger.debug(&quot;shutting down: &quot; &lt;&gt; inspect(reason)) disconnect(socket) end","ref":"Slipstream.html#c:terminate/2","title":"Slipstream.terminate/2","type":"callback"},{"doc":"Any data structure capable of being serialized as JSONAny argument typed as Slipstream.json_serializable/0 must be able to be encoded with the JSON parser passed in configuration. See Slipstream.Configuration.","ref":"Slipstream.html#t:json_serializable/0","title":"Slipstream.json_serializable/0","type":"type"},{"doc":"A reference to a message pushed by the clientThese references are returned by calls to push/4 and may be matched on in handle_reply/3. They are also used to match messages for await_reply/2.SynchronicityThis approach treats the websocket connection as an RPC: some other process in the service does a GenServer.call/3 to the slipstream client process, which sends a push to the remote websocket server, waits for a reply (synchronously) and then sends that back to the caller. All-in-all, this appears completely synchronous for the caller.@impl Slipstream def handle_call({:new_message, params}, _from, socket) do {:ok, ref} = push(socket, &quot;rooms:lobby&quot;, &quot;msg:new&quot;, params) {:reply, await_reply(ref), socket} endThis approach is written in a more asynchronous fashion. An info message arriving from any other process triggers the slipstream client to push a work message to the remote websocket server. When the remote websocket server replies with the result, the slipstream client sends off the result to be dealt with else-where. No process in this scenario blocks, so they are all capable of receiving other messages while the work is being completed.@impl Slipstream def handle_info(:do_work, socket) do ref = push(socket, &quot;worker_queue:foo&quot;, &quot;do_work&quot;, %{}) {:noreply, assign(socket, :work_ref, ref)} end @impl Slipstream def handle_reply(ref, result, %{assigns: %{work_ref: ref}} = socket) do IO.inspect(result, label: &quot;work complete!&quot;) {:ok, socket} end","ref":"Slipstream.html#t:push_reference/0","title":"Slipstream.push_reference/0","type":"type"},{"doc":"A reply from a remote server to a push from the clientReplies may be any of:ok:error{:ok, any()}{:error, any()}depending on how the remote server's reply is written.Note that the empty map is removed in ok and error tuples, so a reply written like so on the server-side:def handle_in(_event, _params, socket) do {:reply, {:ok, %{}}, socket} endwill translate to a reply of :ok (and the same for {:error, %{}}).Examples# on the Phoenix.Channel (server) side: def handle_in(_event, _params, socket) do {:reply, {:ok, %{created?: true}}, socket} end # on the Slipstream (client) side: def handle_reply(_ref, {:ok, %{&quot;created?&quot; =&gt; true}} = _reply, socket) do ..","ref":"Slipstream.html#t:reply/0","title":"Slipstream.reply/0","type":"type"},{"doc":"Configuration for a Slipstream websocket connectionSlipstream server process configuration is passed in with Slipstream.connect/2 (or Slipstream.connect!/2), and so all configuration is evauated and validated at runtime, as opposed to compile-time validation. You should not expect to see validation errors on configuration unless you force the validation at compile-time, e.g.:# you probably don&#39;t want to do this... defmodule MyClient do @config Application.compile_env!(:my_app, __MODULE__) use Slipstream def start_link(args) do Slipstream.start_link(__MODULE__, args, name: __MODULE__) end def init(_args), do: {:ok, connect!(@config)} .. endThis approach will validate the configuration at compile-time, but you will be unable to change the configuration after compilation, so any secrets contained in the configuration (e.g. a basic-auth request header) will be compiled into the beam files.See the docs for Slipstream.init/1 for a safer approach.Options:uri - Required. The endpoint to which the websocket will connect. Schemes of &quot;ws&quot; and &quot;wss&quot; are supported, and a scheme must be provided. Either binaries or URI structs are accepted. E.g. &quot;ws://localhost:4000/socket/websocket&quot;.:heartbeat_interval_msec - The time between heartbeat messages. A value of 0 will disable automatic heartbeat sending. Note that a Phoenix.Channel will close out a connection after 60 seconds of inactivity (60_000). The default value is 30000.:headers - A set of headers to merge with the request headers when GETing the websocket URI. Headers must be provided as two-tuples where both elements are binaries. Casing of these headers is inconsequential. The default value is [].:json_parser - A JSON parser module which exports at least encode/1 and decode/2. The default value is Jason.:reconnect_after_msec - A list of times to reference for trying reconnection when Slipstreamm.reconnect/0 is used to request reconnection. The msec time will be fetched based on its position in the list with Enum.at(reconnect_after_msec, try_number). If the number of tries exceeds the length of the list, the final value will be repeated. The default value is [10, 50, 100, 150, 200, 250, 500, 1000, 2000, 5000].:rejoin_after_msec - A list of times to reference for trying to rejoin a topic when Slipstreamm.rejoin/0 is used. The msec time will be fetched based on its position in the list with Enum.at(rejoin_after_msec, try_number). If the number of tries exceeds the length of the list, the final value will be repeated. The default value is [100, 500, 1000, 2000, 5000, 10000].:gun_open_options - A map of options to pass to :gun.open/3. See the :gun documentation for more information. Note that :gun does not support websocket over HTTP2 and that :gun naively prefers HTTP2 when connecting over TLS. The :protocols =&gt; [:http] option will be merged in by default to allow &quot;wss&quot; connections out of the box. The default value is %{protocols: [:http]}.:test_mode? - Whether or not to start-up the client in test-mode. See Slipstream.SocketTest for notes on testing Slipstream clients. The default value is false.Note that a Phoenix.Channel defined withsocket &quot;/socket&quot;, UserSocket, ..Can be connected to at /socket/websocket.","ref":"Slipstream.Configuration.html","title":"Slipstream.Configuration","type":"module"},{"doc":"Validates a proposed configuration","ref":"Slipstream.Configuration.html#validate/1","title":"Slipstream.Configuration.validate/1","type":"function"},{"doc":"Validates a proposed configuration, raising on error","ref":"Slipstream.Configuration.html#validate!/1","title":"Slipstream.Configuration.validate!/1","type":"function"},{"doc":"","ref":"Slipstream.Configuration.html#t:t/0","title":"Slipstream.Configuration.t/0","type":"type"},{"doc":"A supervisor for connection processesSlipstream.ConnectionSupervisor is a simple module-based DynamicSupervisor which is used to supervise connection processes. As such, you may track the number of connection processes with DynamicSupervisor.count_children/1. Slipstream.ConnectionSupervisor uses its module name as the DynamicSupervisor name, so you may pass Slipstream.ConnectionSupervisor into any DynamicSupervisor function.Examplesiex&gt; Slipstream.ConnectionSupervisor |&gt; DynamicSupervisor.count_children() %{active: 15, specs: 15, supervisors: 0, workers: 15}","ref":"Slipstream.ConnectionSupervisor.html","title":"Slipstream.ConnectionSupervisor","type":"module"},{"doc":"Returns a specification to start this module under a supervisor.See Supervisor.","ref":"Slipstream.ConnectionSupervisor.html#child_spec/1","title":"Slipstream.ConnectionSupervisor.child_spec/1","type":"function"},{"doc":"A data structure representing a potential websocket client connectionThis structure closely resembles Phoenix.Socket.t/0, but is not compatible with its functions. All documented functions from this module are imported by use Slipstream.","ref":"Slipstream.Socket.html","title":"Slipstream.Socket","type":"module"},{"doc":"","ref":"Slipstream.Socket.html#assign/2","title":"Slipstream.Socket.assign/2","type":"function"},{"doc":"Adds key-value pairs to socket assignsBehaves the same as Phoenix.Socket.assign/3Examplesiex&gt; assign(socket, :key, :value) iex&gt; assign(socket, key: :value)","ref":"Slipstream.Socket.html#assign/3","title":"Slipstream.Socket.assign/3","type":"function"},{"doc":"Gets the process ID of the connectionThe slipstream implementor module is not the same process as the GenServer which interfaces with the remote server for websocket communication. This other process, the Slipstream.Connection process, interfaces with :gun and communicates with the implementor module by puassing messages (mostly with Kernel.send/2.It can be useful to have access to this pid for niche purposes, like sending a fake disconnect message or for debugging (e.g. with :sys.get_state/1)Examplesiex&gt; Slipstream.Socket.channel_pid(socket) #PID&lt;0.1.2&gt;","ref":"Slipstream.Socket.html#channel_pid/1","title":"Slipstream.Socket.channel_pid/1","type":"function"},{"doc":"Checks if a socket is connected to a remote websocket hostExamplesiex&gt; socket = connect(socket, uri: &quot;ws://example.org&quot;) iex&gt; socket = await_connect!(socket) iex&gt; connected?(socket) true","ref":"Slipstream.Socket.html#connected?/1","title":"Slipstream.Socket.connected?/1","type":"function"},{"doc":"Checks the status of a join requestWhen a join is requested with Slipstream.join/3, the join request is considered to be in the :requested state. Once the topic is successfully joined, it is considered :joined until closed. If there is a failure to join the topic, if the topic crashes, or if the topic is left after being joined, the status of the join is considered :closed. Finally, if a topic has not been requested in a join so far for a socket, the status is nil.Notably, the status of a join will not automatically change to :joined once the remote server replies with successful join. Either the join must be awaited with Slipstream.await_join/2 or the status may be checked later in the Slipstream.handle_join/3 callback.Examplesiex&gt; socket = join(socket, &quot;room:lobby&quot;) iex&gt; join_status(socket, &quot;room:lobby&quot;) :requested iex&gt; {:ok, socket, _join_response} = await_join(socket, &quot;room:lobby&quot;) iex&gt; join_status(socket, &quot;room:lobby&quot;) :joined","ref":"Slipstream.Socket.html#join_status/2","title":"Slipstream.Socket.join_status/2","type":"function"},{"doc":"Checks if a channel is currently joinedExamplesiex&gt; joined?(socket, &quot;room:lobby&quot;) true","ref":"Slipstream.Socket.html#joined?/2","title":"Slipstream.Socket.joined?/2","type":"function"},{"doc":"A socket data structure representing a potential websocket client connection","ref":"Slipstream.Socket.html#t:t/0","title":"Slipstream.Socket.t/0","type":"type"},{"doc":"Helper functions and macros for testing Slipstream clientsThis module is something of a correlary to Phoenix.ChannelTest. The functions and macros in Phoenix.ChannelTest emulate client operations: functions like Phoenix.ChannelTest.join/2, Phoenix.ChannelTest.push/3, etc.. Functions and macros in this module emulate behavior of the server: a Phoenix.Channel.Timing assumptionsClients are typically written to assume that the serveris up at time of client start-upis always upWhile this is typically (at least mostly) accurate, it is not necessarily true in general and will not be true when testing Slipstream clients. In particular, Slipstream clients may fail in test-mode under the following conditions:the client is started immediately in the application supervision treeit awaits a connection synchronously (with Slipstream.await_connect/2)the testing suite takes longer to complete than the timeout given to Slipstream.await_connect/2While a client that uses Slipstream.await_connect/2 can band-aid over these problems with a high enough timeout, clients should be satisfied with receiving successful connection events asynchronously. Clients written in the asynchronous callback style will not be affected.Some implementation-level details are glossed over by this testing framework, including that of heartbeat timeouts. If a client sits idle after joining for more than 60 seconds, they will not be terminated due to heartbeat timeout.Setting up a testAnother assumption clients typically make of servers is that clients assume there is just one server. A client is not written expecting to hear that it has been connected multiple times for one request to Slipstream.connect/2. As such, tests using this module as a case template should be run synchronously.defmodule MyApp.MyClientTest do use Slipstream.SocketTest ..By default, this will start a server session for each test that simulates the current test process as the websocket server connected to the client.test &quot;the client sends a push to the server on join&quot;, c do accept_connect(MyClient) endThis server does not run a websocket server. Instead the server is a conceptual server: you may imagine that in each test, you are have control of the remote server and can control the behavior of the server imperatively.The assert_* and refute_* family of macros from this module allow you to make assertions about- and match on values from- requests from the client to the server. The remaining functions allow you to emulate actions on behalf of a hypothetical server.TimeoutsThe assert_* and refute_* macros from this module default to ExUnit timeouts. See the ExUnit documentation for more details.FormattingSlipstream exports the assert_* and receive_* macros as valid locals_without_parens in its formatter config. To be able to use these macros without the formatter injecting parentheses, import :slipstream in your service's formatter config for :import_deps:[ inputs: [&quot;mix.exs&quot;, &quot;{config,lib,test}/**/*.{ex,exs}&quot;], import_deps: [:slipstream], .. # more configuration ]","ref":"Slipstream.SocketTest.html","title":"Slipstream.SocketTest","type":"module"},{"doc":"Emulates a server telling the client it has connectedThis sets the current process as the current server connected to the client. It also adds an ExUnit.callbacks.on_exit/2 function that disconnects the client on exit with reason :closed_by_test. To handle this disconnect, clients should match on this pattern in Slipstream.handle_disconnect/2 and either shutdown (in the case that the test controls the spawning of the client, or reconnect, as with Slipstream.reconnect/1. The default implementation of Slipstream.handle_disconnect/2 reconnects in this case.See Timing Assumptions.Examplestest &quot;the server connects&quot;, c do :ok = accept_connect(MyApp.MyClient) end","ref":"Slipstream.SocketTest.html#accept_connect/1","title":"Slipstream.SocketTest.accept_connect/1","type":"function"},{"doc":"Asserts that a client will attempt to disconnect from the serverExamplesaccept_connect(MyClient) # client will disconnect after 15s of inactivity assert_disconnect 15_000","ref":"Slipstream.SocketTest.html#assert_disconnect/1","title":"Slipstream.SocketTest.assert_disconnect/1","type":"macro"},{"doc":"Asserts that a client will request to join a topictopic_expr and params_expr are interpreted as match expressions, so they may be literal values, pinned (^) bindings, or partial values such as &quot;msg:&quot; &lt;&gt; _ or %{} (which matches any map).reply is meant to simulate the return value of the Phoenix.Channel.join/3 callback.Examplesaccept_connect(MyClient) assert_join &quot;rooms:lobby&quot;, %{}, :ok","ref":"Slipstream.SocketTest.html#assert_join/4","title":"Slipstream.SocketTest.assert_join/4","type":"macro"},{"doc":"Asserts that a client will request to leave a topictopic_expr is a pattern, so literal values like &quot;room:lobby&quot; are valid as well as match patterns such as &quot;room:&quot; &lt;&gt; _. Existing bindings must be pinned with the ^ pin operator.Examplestopic = &quot;rooms:lobby&quot; accept_connect(MyClient) assert_join ^topic, %{}, :ok push(MyClient, topic, &quot;leave&quot;, %{}) assert_leave ^topic","ref":"Slipstream.SocketTest.html#assert_leave/2","title":"Slipstream.SocketTest.assert_leave/2","type":"macro"},{"doc":"Asserts that the client will request to push a messsage to the serverNote that topic_expr, event_expr, params_expr, and ref_expr are all pattern expressions. Prior bindings may be used with the ^ pin operator, values may be underscored to ignore, and partial values may be matched (e.g. %{} will match any map).ref_expr can be provided to bind a reference for later use in reply/3.Examplesassert_push &quot;rooms:lobby&quot;, &quot;msg:new&quot;, params, ref reply(MyClient, ref, {:ok, %{status: &quot;ok&quot;, received: params}})","ref":"Slipstream.SocketTest.html#assert_push/5","title":"Slipstream.SocketTest.assert_push/5","type":"macro"},{"doc":"A convenience macro wrapping connection and a join responseThis macro is written for clients that join immediately after a connection has been established, which is a common case.Clients written like so:@impl Slipstream def handle_connect(socket) do {:ok, join(socket, &quot;rooms:lobby&quot;, %{user_id: socket.assigns.user_id})} endMay be tested with connect_and_assert_join/5 as opposed to a separate connect/2 and then an assert_join/5.Examplessocket = connect_and_assert_join MySocketClient, &quot;rooms:lobby&quot;, %{}, :ok push(socket, &quot;rooms:lobby&quot;, &quot;initial-hello&quot;, %{&quot;hello&quot; =&gt; &quot;world&quot;})","ref":"Slipstream.SocketTest.html#connect_and_assert_join/5","title":"Slipstream.SocketTest.connect_and_assert_join/5","type":"macro"},{"doc":"Emulates a server closing a connection to the clientExamplesaccept_connect(MyClient) disconnect(MyClient, :heartbeat_timeout)","ref":"Slipstream.SocketTest.html#disconnect/2","title":"Slipstream.SocketTest.disconnect/2","type":"function"},{"doc":"Emulates a server pushing a message to the clientThis emulates cases of Phoenix.Channel.push/3 by a Phoenix server.Note that this function will not encode or decode the value of params, so conversion from maps of atom-keys to string-keys will not occur as it would when passing messages over-the-wire.Examplestest &quot;the server increments our counter with ping messages&quot;, c do assert Counter.count() == 0 connect_and_assert_join MyClient, &quot;counter-topic&quot;, %{}, :ok push(MyClient, &quot;counter-topic&quot;, &quot;ping&quot;, %{delta: 1}) assert Counter.count() == 1 end","ref":"Slipstream.SocketTest.html#push/4","title":"Slipstream.SocketTest.push/4","type":"function"},{"doc":"Refutes that a client will attempt to disconnect from the serverThe opposite of assert_disconnect/1.Examplesaccept_connect(MyClient) refute_disconnect 10_000","ref":"Slipstream.SocketTest.html#refute_disconnect/1","title":"Slipstream.SocketTest.refute_disconnect/1","type":"macro"},{"doc":"Refutes that a client will request to join a topicThe opposite of assert_join/5.Examplesaccept_connect(MySocketClient) refute_join &quot;rooms:&quot; &lt;&gt; _, %{user_id: 5}","ref":"Slipstream.SocketTest.html#refute_join/3","title":"Slipstream.SocketTest.refute_join/3","type":"macro"},{"doc":"Refutes that a client will request to leave a topicThe opposite of assert_leave/3.Examplesaccept_connect(MyClient) assert_join &quot;rooms:lobby&quot;, %{}, :ok push(MyClient, topic, &quot;no-don&#39;t-go&quot;, %{}) refute_leave ^topic, 10_000","ref":"Slipstream.SocketTest.html#refute_leave/2","title":"Slipstream.SocketTest.refute_leave/2","type":"macro"},{"doc":"Refutes that a client will push a message to the serverThe opposite of assert_push/4Examplesrefute_push &quot;rooms:lobby&quot;, &quot;msg:&quot; &lt;&gt; _, %{user_id: 5}","ref":"Slipstream.SocketTest.html#refute_push/4","title":"Slipstream.SocketTest.refute_push/4","type":"macro"},{"doc":"Emulates a server replying to a push from the clientref is a reference which can be matched upon in assert_push/6. reply follows the Slipstream.reply/0 type: it may be an :ok or :error atom or an {:ok, any()} or {:error, any()} tuple. This value will not be encoded or decoded by Slipstream, instead just directly passed to the client.Examplestopic = &quot;rooms:lobby&quot; connect_and_assert_join MySocketClient, ^topic, %{}, :ok assert_push ^topic, &quot;ping&quot;, %{}, ref reply(MySocketClient, ref, {:ok, %{&quot;ping&quot; =&gt; &quot;pong&quot;}})","ref":"Slipstream.SocketTest.html#reply/3","title":"Slipstream.SocketTest.reply/3","type":"function"},{"doc":"Any Slipstream clientSince Slipstream clients are either GenServer or plain processes, either a pid or a GenServer name will work as the client argument for any function specifying client in this module.","ref":"Slipstream.SocketTest.html#t:client/0","title":"Slipstream.SocketTest.client/0","type":"type"},{"doc":"ChangelogAll notable changes to this project will be documented in this file.The format is based on Keep a Changelog, and this project adheres to Semantic Versioning.","ref":"changelog.html","title":"Changelog","type":"extras"},{"doc":"FixedLink to post-cutover-metrics image made absolute","ref":"changelog.html#0-4-4-2021-02-07","title":"Changelog - 0.4.4 - 2021-02-07","type":"extras"},{"doc":"AddedAdded examples to the documentation&quot;Graceful Startup&quot; and &quot;Repeater-style Client&quot;Added a note about post-cutover performance changes after we (NFIBrokerage) cutover to Slipstream in our stackThis is a new section in the &quot;Why :gun?&quot; guideAdded documentation on counting the number of Slipstream connection processes to Slipstream.ConnectionSupervisor","ref":"changelog.html#0-4-3-2021-02-07","title":"Changelog - 0.4.3 - 2021-02-07","type":"extras"},{"doc":"FixedAdded channel config to socket on Slipstream.connect/2 or Slipstream.connect!/2this fixes an issue reconnecting when faking a connection through Slipstream.SocketTest.accept_connect/1","ref":"changelog.html#0-4-2-2021-02-05","title":"Changelog - 0.4.2 - 2021-02-05","type":"extras"},{"doc":"FixedChanged the default behavior of Slipstream.terminate/2 to disconnect gracefully with Slipstream.disconnect/1AddedAdded better documentation on the Slipstream.terminate/2 callback and how to disconnect in it to leave a connection gracefully","ref":"changelog.html#0-4-1-2021-02-04","title":"Changelog - 0.4.1 - 2021-02-04","type":"extras"},{"doc":"AddedAdded telemetry support for clientssee the telemetry guide for more details","ref":"changelog.html#0-4-0-2021-02-04","title":"Changelog - 0.4.0 - 2021-02-04","type":"extras"},{"doc":"FixedFixed spelling and grammar mistakes in the implementation guide","ref":"changelog.html#0-3-4-2021-02-04","title":"Changelog - 0.3.4 - 2021-02-04","type":"extras"},{"doc":"Addedincreased documentation around retry mechanismsadded a page on the choice of :gun as the low-level webscket client","ref":"changelog.html#0-3-3-2021-02-04","title":"Changelog - 0.3.3 - 2021-02-04","type":"extras"},{"doc":"FixedFixed edge case where the remote server shutting down would not emit a ChannelClosed eventthis and heartbeats caused the odd :gun_error atom from #12","ref":"changelog.html#0-3-2-2021-02-03","title":"Changelog - 0.3.2 - 2021-02-03","type":"extras"},{"doc":"FixedThe telemetry event for message handles in the connection process should now correctly publish the :events key, not the :event keyThe telemetry guide now correctly states that the :start_state key holds the connection state for handle-style eventsAddedPublished the connection state at the end of the message handle for handle-type telemetry events in the connection processthis is published in the :end_state key","ref":"changelog.html#0-3-1-2021-02-03","title":"Changelog - 0.3.1 - 2021-02-03","type":"extras"},{"doc":"Added:telemetry events are executed for each message handled by the Slipstream.Connection processthis should aide in debugging in scenarios in which one wants to see the full event history of a connection","ref":"changelog.html#0-3-0-2021-02-03","title":"Changelog - 0.3.0 - 2021-02-03","type":"extras"},{"doc":"FixedEmit ChannelClosed event on HTTP close from remote websocket serverSee #11","ref":"changelog.html#0-2-4-2021-02-02","title":"Changelog - 0.2.4 - 2021-02-02","type":"extras"},{"doc":"Changeddocs: the main page now shows the Slipstream modulefrankly, I just want people to see the Basic Usage section as soon as possible so they get a taste for the library","ref":"changelog.html#0-2-3-2021-02-02","title":"Changelog - 0.2.3 - 2021-02-02","type":"extras"},{"doc":"FixedThe connection now properly disconnects when the server's heartbeat replies timeout with requestsdisconnects for this reason will send have :heartbeat_timeout as the reason referenced in Slipstream.handle_disconnect/2","ref":"changelog.html#0-2-2-2021-02-01","title":"Changelog - 0.2.2 - 2021-02-01","type":"extras"},{"doc":"FixedConnection now correctly emits ChannelClosed events for :gun events which occur when a service's network access is terminated","ref":"changelog.html#0-2-1-2021-02-01","title":"Changelog - 0.2.1 - 2021-02-01","type":"extras"},{"doc":"AddedAdded a testing framework Slipstream.SocketTest","ref":"changelog.html#0-2-0-2021-02-01","title":"Changelog - 0.2.0 - 2021-02-01","type":"extras"},{"doc":"FixedPublished license to hex","ref":"changelog.html#0-1-4-2021-01-31","title":"Changelog - 0.1.4 - 2021-01-31","type":"extras"},{"doc":"ChangedWidened dependency on :gun from ~&gt; 1.3 to ~&gt; 1.0as far as I can tell, there aren't any real changes to the websocket interface, so any of these versions should be OK","ref":"changelog.html#0-1-3-2021-01-30","title":"Changelog - 0.1.3 - 2021-01-30","type":"extras"},{"doc":"AddedMore documentationincluding a Basic Usage section with a minimalistic example","ref":"changelog.html#0-1-2-2021-01-30","title":"Changelog - 0.1.2 - 2021-01-30","type":"extras"},{"doc":"RemovedRemoved direct dependency on :telemetrysee #4","ref":"changelog.html#0-1-1-2021-01-30","title":"Changelog - 0.1.1 - 2021-01-30","type":"extras"},{"doc":"AddedInitial implementation and initial publish","ref":"changelog.html#0-1-0-2021-01-30","title":"Changelog - 0.1.0 - 2021-01-30","type":"extras"},{"doc":"AddedThis project was generated by Gaas","ref":"changelog.html#0-0-0-2021-01-22","title":"Changelog - 0.0.0 - 2021-01-22","type":"extras"},{"doc":"TelemetrySlipstream emits telemetry events for both processes: the client process and the connection process.Telemetry for the connection process is very low-level: information about each message and before-and-after states for the connection. This telemetry aides in debugging failure-states of Slipstream and can be useful for development, but is likely not worth ingesting for average Slipstream users.Client process telemetry is similar to the telemetry emitted by Phoenix.Channels and is recommended for consumption by average Slipstream users.","ref":"telemetry.html","title":"Telemetry","type":"extras"},{"doc":"Clients emit telemetry in two categories:basic channel connection and topic join eventscallback eventsThe first emulate two-of-three events emitted by Phoenix.Channels (at time of writing):[:slipstream, :client, :connect] - dispatched at the end of a successful client connection to a remote websocket hostmetadata%{ start_time: DateTime.t(), configuration: Slipstream.Configuration.t(), socket: Slipstream.Socket.t(), # emitted only on the stop event: response_headers: [{String.t(), String.t()}] }[:slipstream, :client, :join] - dispatched at the end of a successful join onto a topicmetadata%{ start_time: DateTime.t(), socket: Slipstream.Socket.t(), topic: String.t(), params: Slipstream.json_serializable(), # emitted only on the stop event: response: Slipstream.json_serializable() }The second is emitted for any callback defined by the Slipstream module. Each of these events are emitted under the event name of [:slipstream, :client, callback], where callback is any callback defined by Slipstream, e.g. :handle_message. The metadata for these events is as follows:%{ client: module(), callback: atom(), arguments: [any()], socket: Slipstream.Socket.t(), start_time: DateTime.t(), # emitted only on the stop event response: any() }Note that all client-process events emulate :telemetry.span/3 in naming and in measurements, and in the case of callback events, the events are emitted by :telemetry.span/3. This means that each of the above-described events are actually prefixes, and each event name below represents messages of event_prefix ++ [:start] and event_prefix ++ [:stop], with exceptions being emitted as event_prefix ++ [:exception].Client authors are encouraged to only subscribe to the callbacks they are interested in (e.g. :handle_message or :handle_reply).Note: events will not be emitted for any messages handled by the await_* family of synchronous functions.","ref":"telemetry.html#client-process-telemetry","title":"Telemetry - Client-process Telemetry","type":"extras"},{"doc":"Slipstream currently emits telemetry for each message received in its Slipstream.Connection process. This aides in low-level debugging when one desires to see the entire event-history of a connection.A connection can be pieced together by the :trace_id and :connection_id keys emitted in the metadata of these telemetry events. One Slipstream.Connection process is guaranteed to only use one :trace_id and one :connection_id.The trace-id acts as a identifier which groups together all requests belonging to that Slipstream.Connection process. Because the life-cycle of the GenServer which runs the Slipstream.Connection process and the actual connection to a remote websocket are the same, there doesn't appear to be much difference between the :trace_id and the :connection_id, however, the :connection_id can be considered the parent-ID for all spans, where each subsequent message to the connection process is considered a span.The Slipstream.Connection telemetry event prefixes are as follows:[:slipstream, :connection, :connect]: an event which is executed at the start and stop of the connection process's life-cycle.The metadata contains:%{ state: %Slipstream.Connection.State{}, connection_id: String.t(), trace_id: String.t(), # UTC date-time stamp of start-of-connection start_time: DateTime.t() }[:slipstream, :connection, :handle]: an event executed for each message received by the connection process.Metadata contains:%{ start_state: %Slipstream.Connection.State{}, connection_id: String.t(), trace_id: String.t(), # unique to this message handle span_id: String.t(), # UTC date-time stamp of start-of-handling start_time: DateTime.t(), raw_message: term(), # included on the `event_prefix ++ [:stop]` events: message: term(), events: [%{type: atom(), attrs: map()}], built_events: [struct()], return: tuple(), end_state: %Slipstream.Connection.State{} }Each of these events follows the :telemetry.span/3 pattern in measurements and in names: event_prefix ++ [:start] and event_prefix ++ [:stop] events are emitted using each of the above event prefixes.Note: connection telemetry will not be emitted for any connections simulated in a Slipstream.SocketTest.","ref":"telemetry.html#connection-process-telemetry","title":"Telemetry - Connection-process Telemetry","type":"extras"},{"doc":"ImplementationSlipstream has some implementation decisions that may appear strange or overly-complicated on the surface. My opinion is that these help not only with usability but also test-ability. The test suite is rather compact: it covers a great many cases for the small number of lines in which it is implemented.The primary design decisions that enable this easy workflow are the hard boundary between the client and connection processes and the disambiguation of messages into events and commands:","ref":"implementation.html","title":"Implementation","type":"extras"},{"doc":"Slipstream starts two processes to run a websocket connection:the client process (modules with use Slipstream)the connection processThe connection process is spawned and killed upon Slipstream.connect/2 and Slipstream.disconnect/1, respectively. It directly interfaces with :gun in active mode, meaning that messages from :gun connections and requests are forwarded to the mailbox of the connection process.The client process is simply a GenServer that has GenServer.handle_info/2 clauses injected to handle slipstream events. The entire purpose of the client process is to provide a clean-feeling interface to the connection process. One could (but is not recommended to-) forego the client process and deal with websockets synchronously with the Slipstream.await_* family of functions.","ref":"implementation.html#process-architecture","title":"Implementation - Process Architecture","type":"extras"},{"doc":"Slipstream separates out messages into two categories: events and commands. To some extent, Slipstream follows the Event Sourcing and Command Pattern design patterns with this choice.One might think to implement the communication between the connection and client processes by using data structures like %Message{} and %Reply{} and %Heartbeat{}, but it's not exactly clear who should send which message to whom. Should the connection process be sending %Message{} to the server, or vice-versa? Could we re-use that data structure for both directions?Slipstream's Slipstream.Events and Slipstream.Commands data structures keep this boundary clear. The client process sends commands to the connection process when it seeks to change the connection state (join/leave) or push a message. The connection process notifies the client of its status and how the connection changes (whenever it changes meaningfully), with events. Events declare how the connection has changed, which informs the client process of how it should change the state of the Slipstream.Socket.t/0 as the connection evolves.Events are written in the past tense in the form of NounVerbed, as in ThingHappened, while commands are written in the form of VerbNoun, as in DoThing.SignaturesEvents arrive from the connection process in the format of{:__slipstream_event__, event_struct}where event_struct is any struct from Slipstream.Events. And commands are sent to the connection process in the form of:{:__slipstream_command__, command_struct}The Slipstream.Signatures file defines two macros that wrap the above patterns. This allows one to both produce and match on a data structure and have confidence that the resulting data structure or match pattern will be an event or command.This pattern is not very revolutionary: it's really just an extension of DRY into match patterns. It keeps the matches and routing functions clear of dirty implementation details like 'exactly how the data structures are wrapped.'For example, a HeartbeatAcknowledged event can be produced like so:alias Slipstream.Events import Slipstream.Signatures, only: [event: 1] event(%Events.HeartbeatAcknowledged{}) #=&gt; {:__slipstream_event__, %Slipstream.Events.HeartbeatAcknowledged{}}And then later-on this data structure can be matched with any sort of clause like so:alias Slipstream.Events import Slipstream.Signatures, only: [event: 1] maybe_event = event(%Events.HeartbeatAcknowledged{}) case maybe_event do event(%Events.HeartbeatAcknowledged{} = event) -&gt; {:ok, event} _ -&gt; :error end #=&gt; {:ok, %Events.HeartbeatAcknowledged{}}Since the client and connection process communicate with each other though message sending and GenServer.call/3, these signature macros keep the messages recognizable and enforce the communication boundary.I dub this pattern the &quot;signature pattern&quot; (but realize that I am probably not the first to discover it).","ref":"implementation.html#events-and-commands","title":"Implementation - Events and Commands","type":"extras"},{"doc":"The Slipstream.Socket.t/0 data structure is built to closely resemble the Phoenix.Socket.t/0 structure, and feel similar in usage. These are both implementations of a pattern coined by @rrrene: the token pattern.The idea is fairly straight-forward: instead of passing around only the values each function needs, you pass around a structure which contains all information available and provide a functional API to modify the internals (by producing a new structure). These structures are called &quot;tokens.&quot;The token pattern is particularly useful for writing flexible pipelines. E.g. a piece of code that does something like this:def do_work(request) do request |&gt; decode_request() |&gt; hydrate_request_with_background_info() |&gt; start_computation_measurements() |&gt; do_the_computation() |&gt; end_computation_measurements() |&gt; send_the_result_somewhere() endThese functions are all chained together, so it may seem perfectly appropriate to write the functions in a way that each consumes the direct output of the preceding function in the chain. But if we eventually need another do_work/1 clause that handles a different sort of request and say, cuts out the hydrate_request_with_background_info/1 step, we will likely have to refactor at least a few of the functions to properly emit and consume the data they need.Instead of taking and giving only the information each function needs, each function should pass along a token structure containing all the information in the request:def do_work(request) do %Token{initial_request: request} |&gt; decode_request() # returns a new %Token{} |&gt; hydrate_request_with_background_info() # returns a new %Token{} .. endIf we need to cut out or add new functions to the pipeline, we are less likely to need to refactor any of the component functions.The usage of sockets in Slipstream is not so chain-y as the example, but the socket token allows a clean API of functions that each take a socket and some other information and emit a new socket token.@impl Slipstream def handle_info({:join, topic}, socket) do socket = if connected?(socket) and not joined?(socket, topic) do join(socket, topic) else socket end {:noreply, socket} end","ref":"implementation.html#the-token-pattern","title":"Implementation - The Token Pattern","type":"extras"},{"doc":"Why :gun?Slipstream is built off of :gun, an HTTP and websocket client from the wonderful @ninenines family of projects. (Others include :cowboy and :ranch, which power Phoenix.):gun is battle-tested and fun to use, but the main feature we wanted gun for is that it is not :websocket_client.Prior to Slipstream, we at NFIBrokerage used phoenix_gen_socket_client and phoenix_client, which are both backed by :websocket_client. Once we started using phoenix channel clients in our production services en masse, we found an odd bug we believe to be a known issue in :websocket_client around the RFC-conformance of close frames. This closing issue would manifest itself as zombie procesess on the server-side (specifically, :cowboy_clear.connection_process/4 processes) which would never release process memory.This isn't likely a common problem for most phoenix websocket client users, but at the time we were initially testing out a framework for fueling front-ends and keeping them up-to-date with a library called Haste, which we intend on open-sourcing soon. Haste uses websockets rather aggressively, connecting and disconnecting them very quickly and moving large amounts of data over-the-wire.In addition to loading pages of data over-the-wire on the websocket connection, the Phoenix.Channel on the server-side would attempt to determine if changes made in the database would affect rows on a users table. This resulted in a large numbers of messages being sent to the Phoenix.Channel, which prevented process hibernation, leading to a very large process memory. The stacking up of this process memory and the zombie-like state these processes were left in when :websocket_client attempted to disconnect would quickly overwhelm the backend server hosting the data, leading to out-of-memory crashes.","ref":"why_gun.html","title":"Why :gun?","type":"extras"},{"doc":"We cutover to the majority of our phoenix socket clients to Slipstream at ~13:29 on 02/02, seen as the black bar in the graph above.Note that disconnection reasons such as &quot;LiveView process terminated&quot;, all &quot;parent process terminated&quot; and &quot;subscription process terminated (cleanup)&quot; are benign exit reasons.Once we cutover our sockets to Slipstream, we found a marked decline in clients terminating due to remote server disconnect. It's hard to point the finger at the exact reason this decline took place, but it is likely the product of a less-encumbered backend.","ref":"why_gun.html#post-cutover-performance","title":"Why :gun? - Post-Cutover Performance","type":"extras"},{"doc":"ExamplesSlipstream provides a number of example socket client implementations as both a tutorial-like guide and a reference for common implementation patterns.Each of these examples is tested (via mix test), so each example is assured to be in working order by every green check-mark from the CI.","ref":"examples.html","title":"Examples","type":"extras"},{"doc":"Each example is contained in a directory in ./examples. Each example contains a README.md with:a description of the example and what it tries to solvea tutorial-like listing of commits that created the example, along with explanations of each stepIn order to use the examples as a tutorial, try reading through each commit and reproducing the client in an example project.","ref":"examples.html#format","title":"Examples - Format","type":"extras"},{"doc":"Graceful Startup: a client that handles failures in configuration at start-up time with graceful degradationRepeater-style Client: a client which subscribes to a topic in another service in order to re-publish each message as a broadcast in the client's service's Phoenix.Endpoint","ref":"examples.html#directory","title":"Examples - Directory","type":"extras"},{"doc":"Graceful StartupA client that gracefully handles errors in configuration at start-up time.As the documentation for GenServer.init/1 details, a GenServer may return :ignore instead of {:ok, initial_state} in its GenServer.init/1 callback. This will cause the client toexit normally without entering the loop or calling GenServer.terminate/2. If used when part of a supervision tree the parent supervisor will not fail to start nor immediately try to restart the GenServer.This makes it a good option for handling failures to start the client. Slipstream.init/1 is mostly a wrapper around GenServer.init/1, so we may return :ignore and see the same behavior.","ref":"graceful_startup_example.html","title":"Graceful Startup","type":"extras"},{"doc":"We start with an empty client module (c5a08fd)defmodule MyApp.GracefulStartupClient do endAnd immediately fill out the Slipstream basics like a start_link/1 implementation (so the module may be supervised), and an invocation of use Slipstream (8d44dd0)defmodule MyApp.GracefulStartupClient do use Slipstream @moduledoc &quot;...&quot; def start_link(opts) do Slipstream.start_link(__MODULE__, opts, name: __MODULE__) end endThen we add a basic implementation of the Slipstream.init/1 callback (a51e77c)@impl Slipstream def init(_args) do config = Application.fetch_env!(:my_app, __MODULE__) {:ok, connect!(config)} endBut note that this introduces two potential raiseing errors in our Slipstream.init/1 callback:Application.fetch_env!/2 will raise of the key-value pair is not defined in configuration (config/*.exs)Slipstream.connect!/2 will raise if the configuration passed as the first argument is not valid according to Slipstream.Configuration.And also note that a raise in an Slipstream.init/1 callback will fail the start-up of the supervisor process. If this client is started in the Application supervision tree (lib/my_app/application.ex), it will take down the entire application on error.So let's refactor this to make it a bit more safe! First, we switch Application.fetch_env!/2 to its more graceful counterpart: Application.fetch_env/2, which returns {:ok, config} when the configuration is defined and :error when it is not (90ceb4a)@impl Slipstream def init(_args) do with {:ok, config} &lt;- Application.fetch_env(:slipstream, __MODULE__) do {:ok, connect!(config)} else :error -&gt; :ignore end endNow the client will attempt to connect only if the configuration is defined. But it can still fail, as we use the raising Slipstream.connect!/2. We refactor that to Slipstream.connect/2 in b3be445:@impl Slipstream def init(_args) do with {:ok, config} &lt;- Application.fetch_env(:slipstream, __MODULE__), {:ok, socket} &lt;- connect(config) do {:ok, socket} else :error -&gt; :ignore {:error, _reason} -&gt; :ignore end endSo now in either of our failure-cases, the client will simply not start-up instead of potentially crashing the entire application. This is an application of graceful degradation: in cases of system failure, we degrade our performance instead of entirely giving up.In this example, we go on to add helpful Logger messages that declare when a failure case has been met (09d1a83). To see the full example code, open up examples/graceful_startup/client.ex. A small test suite can be found at test/slipstream/examples/graceful_startup_test.exs.","ref":"graceful_startup_example.html#tutorial","title":"Graceful Startup - Tutorial","type":"extras"},{"doc":"Repeater-style ClientWe call any client that repeats messages from another service's endpoint a repeater client.This is a useful pattern when one wishes to ferry messages (without changing them) from one Phoenix.Endpoint to another when the services are different, or when the services are the same but endpoint clustering is not possible or desired.In order to make this happen, we will create a client thatconnects to the remove servicesubscribes to the topic we wish to repeatbroadcasts messages as they happen to this service's endpoint","ref":"repeater_example.html","title":"Repeater-style Client","type":"extras"},{"doc":"We start with an empty client (db79a32)defmodule MyApp.RepeaterClient do @moduledoc &quot;&quot;&quot; A repeater-kind of client which re-broadcasts messages from another service into this service&#39;s endpoint &quot;&quot;&quot; use Slipstream def start_link(config) do Slipstream.start_link(__MODULE__, config, name: __MODULE__) end @impl Slipstream def init(config), do: {:ok, connect!(config)} endThis client is pretty standard. It receives configuration from its supervisor, so it could be started in a lib/my_app/application.ex file with something like:def MyApp.Application do use Application def start(_type, _args) do children = [ {MyApp.RepeaterClient, uri: &quot;ws://service-b.local:4000/socket/websocket&quot;}, MyAppWeb.Endpoint ] Supervisor.start_link(children, strategy: :one_for_one) end endFirst thing's first, let's make sure we join our desired topic upon connection by adding a Slipstream.handle_connect/1 implementation that calls Slipstream.join/3 (37b392e)@topic &quot;rooms:lobby&quot; @impl Slipstream def handle_connect(socket), do: {:ok, join(socket, @topic)}And let's add an empty implementation of Slipstream.handle_message/4 (3855594)@impl Slipstream def handle_message(topic, event, payload, socket) do {:ok, socket} endSlipstream.handle_message/4 is invoked whenever the remote server performs a Phoenix.Channel.push/3, or whenever the topic to which the client is subscribed receives a broadcast. Broadcasts are sent to all subscribers of a topic without a Phoenix.Channel explicitly needing to use Phoenix.Channel.push/3, but the interface is the same for Slipstream clients: Slipstream.handle_message/4 will be invoked with the broadcast.Now for the &quot;repeater&quot; part, we will publish these broadcasts or pushes onto our service's endpoint using Phoenix.Endpoint.broadcast/3 (01b5568)@endpoint SlipstreamWeb.Endpoint @impl Slipstream def handle_message(topic, event, payload, socket) do @endpoint.broadcast(topic, event, payload) {:ok, socket} endAnd that's all! Each Phoenix.Channel.push/3 or Phoenix.Endpoint.broadcast/3 published on the topic will be (re)broadcasted on the endpoint in the service of the repeater. See examples/repeater/client.ex for the full code of this example client and a small test case at test/slipstream/examples/repeater_test.exs.","ref":"repeater_example.html#tutorial","title":"Repeater-style Client - Tutorial","type":"extras"}]